import base64
from urllib.parse import urlencode, unquote
from typing import Dict, Any, TYPE_CHECKING

from mangum.types import QueryParams

from .abstract_handler import AbstractHandler
from .. import Response, Request


if TYPE_CHECKING:  # pragma: no cover
    from awslambdaric.lambda_context import LambdaContext


class AwsApiGateway(AbstractHandler):
    """
    Handles AWS API Gateway events, transforming them into ASGI Scope and handling
    responses

    See: https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway.html
    """

    TYPE = "AWS_API_GATEWAY"

    def __init__(
        self,
        trigger_event: Dict[str, Any],
        trigger_context: "LambdaContext",
        api_gateway_base_path: str,
    ):
        super().__init__(trigger_event, trigger_context)
        self.api_gateway_base_path = api_gateway_base_path

    @property
    def request(self) -> Request:
        event = self.trigger_event

        # See this for more info on headers:
        # https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html#apigateway-multivalue-headers-and-parameters
        headers = {}
        # Read headers
        if event.get("headers"):
            headers.update({k.lower(): v for k, v in event.get("headers", {}).items()})
        # Read multiValueHeaders
        # This overrides headers that have the same name
        # That means that multiValue versions of headers take precedence
        # over their plain versions
        if event.get("multiValueHeaders"):
            headers.update(
                {
                    k.lower(): ", ".join(v) if isinstance(v, list) else ""
                    for k, v in event.get("multiValueHeaders", {}).items()
                }
            )

        request_context = event["requestContext"]

        source_ip = request_context.get("identity", {}).get("sourceIp")
        path = unquote(self._strip_base_path(event["path"])) if event["path"] else "/"
        http_method = event["httpMethod"]
        query_string = self._encode_query_string()

        server_name = headers.get("host", "mangum")
        if ":" not in server_name:
            server_port = headers.get("x-forwarded-port", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))
        client = (source_ip, 0)

        return Request(
            method=http_method,
            headers=[[k.encode(), v.encode()] for k, v in headers.items()],
            path=path,
            scheme=headers.get("x-forwarded-proto", "https"),
            query_string=query_string,
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    def _encode_query_string(self) -> bytes:
        """
        Encodes the queryStringParameters.
        """

        params: QueryParams = self.trigger_event.get(
            "multiValueQueryStringParameters", {}
        )
        if not params:
            params = self.trigger_event.get("queryStringParameters", {})
        if not params:
            return b""
        return urlencode(params, doseq=True).encode()

    def _strip_base_path(self, path: str) -> str:
        if self.api_gateway_base_path and self.api_gateway_base_path != "/":
            if not self.api_gateway_base_path.startswith("/"):
                self.api_gateway_base_path = f"/{self.api_gateway_base_path}"
            if path.startswith(self.api_gateway_base_path):
                path = path[len(self.api_gateway_base_path) :]
        return path

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        headers, multi_value_headers = self._handle_multi_value_headers(
            response.headers
        )

        body, is_base64_encoded = self._handle_base64_response_body(
            response.body, headers
        )

        return {
            "statusCode": response.status,
            "headers": headers,
            "multiValueHeaders": multi_value_headers,
            "body": body,
            "isBase64Encoded": is_base64_encoded,
        }



import base64
import urllib.parse
from typing import Dict, Any, List, Tuple

from . import AwsApiGateway
from .. import Response, Request


class AwsHttpGateway(AwsApiGateway):
    """
    Handles AWS HTTP Gateway events (v1.0 and v2.0), transforming them into ASGI Scope
    and handling responses

    See: https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html  # noqa: E501
    """

    TYPE = "AWS_HTTP_GATEWAY"

    @property
    def event_version(self) -> str:
        return self.trigger_event.get("version", "")

    @property
    def request(self) -> Request:
        event = self.trigger_event

        headers = {}
        if event.get("headers"):
            headers = {k.lower(): v for k, v in event.get("headers", {}).items()}

        request_context = event["requestContext"]

        # API Gateway v2
        if self.event_version == "2.0":
            source_ip = request_context["http"]["sourceIp"]
            path = request_context["http"]["path"]
            http_method = request_context["http"]["method"]
            query_string = event.get("rawQueryString", "").encode()

            if event.get("cookies"):
                headers["cookie"] = "; ".join(event.get("cookies", []))

        # API Gateway v1
        elif self.event_version == "1.0":
            # v1.0 of the HTTP Gateway supports multiValueHeaders
            if event.get("multiValueHeaders"):
                headers.update(
                    {
                        k.lower(): ", ".join(v) if isinstance(v, list) else ""
                        for k, v in event.get("multiValueHeaders", {}).items()
                    }
                )

            source_ip = request_context.get("identity", {}).get("sourceIp")
            path = event["path"]
            http_method = event["httpMethod"]
            query_string = self._encode_query_string()
        else:
            raise RuntimeError(
                "Unsupported version of HTTP Gateway Spec, only v1.0 and v2.0 are "
                "supported."
            )

        server_name = headers.get("host", "mangum")
        if ":" not in server_name:
            server_port = headers.get("x-forwarded-port", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))
        client = (source_ip, 0)

        if not path:
            path = "/"
        else:
            path = self._strip_base_path(path)

        return Request(
            method=http_method,
            headers=[[k.encode(), v.encode()] for k, v in headers.items()],
            path=urllib.parse.unquote(path),
            scheme=headers.get("x-forwarded-proto", "https"),
            query_string=query_string,
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:
        """
        This handles some unnecessary magic from AWS

        >  API Gateway can infer the response format for you
        Boooooo

        https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html#http-api-develop-integrations-lambda.response
        """
        if self.event_version == "1.0":
            return self.transform_response_v1(response)
        elif self.event_version == "2.0":
            return self.transform_response_v2(response)
        raise RuntimeError(  # pragma: no cover
            "Misconfigured event unable to return value, unsupported version."
        )

    def transform_response_v1(self, response: Response) -> Dict[str, Any]:
        headers, multi_value_headers = self._handle_multi_value_headers(
            response.headers
        )

        body, is_base64_encoded = self._handle_base64_response_body(
            response.body, headers
        )
        return {
            "statusCode": response.status,
            "headers": headers,
            "multiValueHeaders": multi_value_headers,
            "body": body,
            "isBase64Encoded": is_base64_encoded,
        }

    def _combine_headers_v2(
        self, input_headers: List[List[bytes]]
    ) -> Tuple[Dict[str, str], List[str]]:
        output_headers: Dict[str, str] = {}
        cookies: List[str] = []
        for key, value in input_headers:
            normalized_key: str = key.decode().lower()
            normalized_value: str = value.decode()
            if normalized_key == "set-cookie":
                cookies.append(normalized_value)
            else:
                if normalized_key in output_headers:
                    normalized_value = (
                        f"{output_headers[normalized_key]},{normalized_value}"
                    )
                output_headers[normalized_key] = normalized_value
        return output_headers, cookies

    def transform_response_v2(self, response_in: Response) -> Dict[str, Any]:
        # The API Gateway will infer stuff for us, but we'll just do that inference
        # here and keep the output consistent

        headers, cookies = self._combine_headers_v2(response_in.headers)

        if "content-type" not in headers and response_in.body is not None:
            headers["content-type"] = "application/json"

        body, is_base64_encoded = self._handle_base64_response_body(
            response_in.body, headers
        )
        response_out = {
            "statusCode": response_in.status,
            "body": body,
            "headers": headers or None,
            "cookies": cookies or None,
            "isBase64Encoded": is_base64_encoded,
        }
        return {key: value for key, value in response_out.items() if value is not None}



import os
import logging
from typing import Any
from urllib.parse import ParseResult, urlparse, parse_qs

import aioboto3
from botocore.exceptions import ClientError

from .base import WebSocketBackend
from ..exceptions import WebSocketError

logger = logging.getLogger("mangum.backends.s3")


def get_file_key(parsed_dsn: ParseResult) -> str:
    if parsed_dsn.path and parsed_dsn.path != "/":
        if not parsed_dsn.path.endswith("/"):
            return parsed_dsn.path + "/"
        return parsed_dsn.path

    return ""


class S3Backend(WebSocketBackend):
    async def __aenter__(self) -> WebSocketBackend:
        parsed_dsn = urlparse(self.dsn)
        parsed_query = parse_qs(parsed_dsn.query)
        self.bucket = parsed_dsn.hostname
        self.key = get_file_key(parsed_dsn)
        self.region_name = (
            parsed_query["region"][0]
            if "region" in parsed_query
            else os.environ["AWS_REGION"]
        )
        self.endpoint_url = (
            parsed_query["endpoint_url"][0] if "endpoint_url" in parsed_query else None
        )

        session = aioboto3.Session()
        self.client = await session.client(
            "s3",
            region_name=self.region_name,
            endpoint_url=self.endpoint_url,
        ).__aenter__()

        create_bucket = False

        try:
            await self.client.head_bucket(Bucket=self.bucket)
        except ClientError as exc:
            error_code = int(exc.response["Error"]["Code"])
            if error_code == 403:  # pragma: no cover
                logger.error("S3 bucket access forbidden!")
            elif error_code == 404:
                logger.info(f"Bucket {self.bucket} not found, creating.")
                create_bucket = True

        if create_bucket:
            await self.client.create_bucket(Bucket=self.bucket)

        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        await self.client.__aexit__(*exc_info)

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        await self.client.put_object(
            Body=json_scope.encode(),
            Bucket=self.bucket,
            Key=f"{self.key}{connection_id}",
        )

    async def retrieve(self, connection_id: str) -> str:
        try:
            s3_object = await self.client.get_object(
                Bucket=self.bucket, Key=f"{self.key}{connection_id}"
            )
        except self.client.exceptions.NoSuchKey:
            raise WebSocketError(f"Connection not found: {connection_id}")

        async with s3_object["Body"] as body:
            scope = await body.read()
            json_scope = scope.decode()

        return json_scope

    async def delete(self, connection_id: str) -> None:
        await self.client.delete_object(
            Bucket=self.bucket, Key=f"{self.key}{connection_id}"
        )



import os
from typing import Any
from urllib.parse import ParseResult, urlparse, parse_qs
import logging

import aioboto3
from botocore.exceptions import ClientError

from .base import WebSocketBackend
from ..exceptions import WebSocketError

logger = logging.getLogger("mangum.backends.dynamodb")


def get_table_name(parsed_dsn: ParseResult) -> str:
    netloc = parsed_dsn.netloc
    _, _, hostinfo = netloc.rpartition("@")
    hostname, _, _ = hostinfo.partition(":")
    return hostname


class DynamoDBBackend(WebSocketBackend):
    async def __aenter__(self) -> WebSocketBackend:
        parsed_dsn = urlparse(self.dsn)
        parsed_query = parse_qs(parsed_dsn.query)
        self.table_name = get_table_name(parsed_dsn)
        self.region_name = (
            parsed_query["region"][0]
            if "region" in parsed_query
            else os.environ["AWS_REGION"]
        )
        self.endpoint_url = (
            parsed_query["endpoint_url"][0] if "endpoint_url" in parsed_query else None
        )

        session = aioboto3.Session()
        self.resource = await session.resource(
            "dynamodb",
            region_name=self.region_name,
            endpoint_url=self.endpoint_url,
        ).__aenter__()

        create_table = False

        try:
            await self.resource.meta.client.describe_table(TableName=self.table_name)
        except ClientError as exc:
            if exc.response["Error"]["Code"] == "ResourceNotFoundException":
                logger.info(f"Table {self.table_name} not found, creating.")
                create_table = True
            else:  # pragma: no cover
                await self.__aexit__(None, None, None)
                raise WebSocketError(exc)

        self.table = await self.resource.Table(self.table_name)

        if create_table:
            client = self.resource.meta.client
            await client.create_table(
                TableName=self.table_name,
                KeySchema=[{"AttributeName": "connectionId", "KeyType": "HASH"}],
                AttributeDefinitions=[
                    {"AttributeName": "connectionId", "AttributeType": "S"}
                ],
                ProvisionedThroughput={
                    "ReadCapacityUnits": 5,
                    "WriteCapacityUnits": 5,
                },
            )
            await self.table.wait_until_exists()

        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        await self.resource.__aexit__(*exc_info)

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        await self.table.put_item(
            Item={"connectionId": connection_id, "initial_scope": json_scope},
            ConditionExpression="attribute_not_exists(connectionId)",
        )

    async def retrieve(self, connection_id: str) -> str:
        try:
            response = await self.table.get_item(Key={"connectionId": connection_id})
            item = response["Item"]
        except KeyError:
            raise WebSocketError(f"Connection not found: {connection_id}")
        initial_scope = item["initial_scope"]
        return initial_scope

    async def delete(self, connection_id: str) -> None:
        await self.table.delete_item(Key={"connectionId": connection_id})



from typing import Any

import aiopg

from .base import WebSocketBackend
from ..exceptions import WebSocketError


class PostgreSQLBackend(WebSocketBackend):
    async def __aenter__(self) -> WebSocketBackend:
        self.connection = await aiopg.connect(self.dsn)
        self.cursor = await self.connection.cursor()

        await self.cursor.execute(
            "create table if not exists mangum_websockets "
            "(id varchar(64) primary key, initial_scope text)"
        )

        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        self.cursor.close()
        await self.connection.close()

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        await self.cursor.execute(
            "insert into mangum_websockets values (%s, %s)",
            (connection_id, json_scope),
        )

    async def retrieve(self, connection_id: str) -> str:
        await self.cursor.execute(
            "select initial_scope from mangum_websockets where id = %s",
            (connection_id,),
        )
        row = await self.cursor.fetchone()
        if not row:
            raise WebSocketError(f"Connection not found: {connection_id}")
        initial_scope = row[0]
        return initial_scope

    async def delete(self, connection_id: str) -> None:
        await self.cursor.execute(
            "delete from mangum_websockets where id = %s", (connection_id,)
        )



import base64
from urllib.parse import urlencode, unquote, unquote_plus
from typing import Any, Dict, Generator, List, Tuple
from itertools import islice

from mangum.types import QueryParams

from .abstract_handler import AbstractHandler
from .. import Response, Request


def all_casings(input_string: str) -> Generator[str, None, None]:
    """
    Permute all casings of a given string.
    A pretty algoritm, via @Amber
    http://stackoverflow.com/questions/6792803/finding-all-possible-case-permutations-in-python
    """
    if not input_string:
        yield ""
    else:
        first = input_string[:1]
        if first.lower() == first.upper():
            for sub_casing in all_casings(input_string[1:]):
                yield first + sub_casing
        else:
            for sub_casing in all_casings(input_string[1:]):
                yield first.lower() + sub_casing
                yield first.upper() + sub_casing


def case_mutated_headers(multi_value_headers: Dict[str, List[str]]) -> Dict[str, str]:
    """Create str/str key/value headers, with duplicate keys case mutated."""
    headers: Dict[str, str] = {}
    for key, values in multi_value_headers.items():
        if len(values) > 0:
            casings = list(islice(all_casings(key), len(values)))
            for value, cased_key in zip(values, casings):
                headers[cased_key] = value
    return headers


class AwsAlb(AbstractHandler):
    """
    Handles AWS Elastic Load Balancer, really Application Load Balancer events
    transforming them into ASGI Scope and handling responses

    See:
        1. https://docs.aws.amazon.com/lambda/latest/dg/services-alb.html
        2. https://docs.aws.amazon.com/elasticloadbalancing/latest/application/lambda-functions.html  # noqa: E501
    """

    TYPE = "AWS_ALB"

    def _encode_query_string(self) -> bytes:
        """
        Encodes the queryStringParameters.
        The parameters must be decoded, and then encoded again to prevent double
        encoding.

        https://docs.aws.amazon.com/elasticloadbalancing/latest/application/lambda-functions.html  # noqa: E501
        "If the query parameters are URL-encoded, the load balancer does not decode
        them. You must decode them in your Lambda function."

        Issue: https://github.com/jordaneremieff/mangum/issues/178
        """

        params: QueryParams = self.trigger_event.get(
            "multiValueQueryStringParameters", {}
        )
        if not params:
            params = self.trigger_event.get("queryStringParameters", {})
        if not params:
            return b""
        params = {
            unquote_plus(key): unquote_plus(value)
            if isinstance(value, str)
            else tuple(unquote_plus(element) for element in value)
            for key, value in params.items()
        }
        return urlencode(params, doseq=True).encode()

    def transform_headers(self) -> List[Tuple[bytes, bytes]]:
        """Convert headers to a list of two-tuples per ASGI spec.

        Only one of `multiValueHeaders` or `headers` should be defined in the
        trigger event. However, we act as though they both might exist and pull
        headers out of both.
        """
        headers: List[Tuple[bytes, bytes]] = []
        if "multiValueHeaders" in self.trigger_event:
            for k, v in self.trigger_event["multiValueHeaders"].items():
                for inner_v in v:
                    headers.append((k.lower().encode(), inner_v.encode()))
        else:
            for k, v in self.trigger_event["headers"].items():
                headers.append((k.lower().encode(), v.encode()))
        return headers

    @property
    def request(self) -> Request:
        event = self.trigger_event

        headers = self.transform_headers()
        list_headers = [list(x) for x in headers]
        # Unique headers. If there are duplicates, it will use the last defined.
        uq_headers = {k.decode(): v.decode() for k, v in headers}

        source_ip = uq_headers.get("x-forwarded-for", "")
        path = unquote(event["path"]) if event["path"] else "/"
        http_method = event["httpMethod"]
        query_string = self._encode_query_string()

        server_name = uq_headers.get("host", "mangum")
        if ":" not in server_name:
            server_port = uq_headers.get("x-forwarded-port", 80)
        else:
            server_name, server_port = server_name.split(":")  # pragma: no cover
        server = (server_name, int(server_port))
        client = (source_ip, 0)

        return Request(
            method=http_method,
            headers=list_headers,
            path=path,
            scheme=uq_headers.get("x-forwarded-proto", "https"),
            query_string=query_string,
            server=server,
            client=client,
            trigger_event=self.trigger_event,
            trigger_context=self.trigger_context,
            event_type=self.TYPE,
        )

    @property
    def body(self) -> bytes:
        body = self.trigger_event.get("body", b"") or b""

        if self.trigger_event.get("isBase64Encoded", False):
            return base64.b64decode(body)
        if not isinstance(body, bytes):
            body = body.encode()

        return body

    def transform_response(self, response: Response) -> Dict[str, Any]:

        multi_value_headers: Dict[str, List[str]] = {}
        for key, value in response.headers:
            lower_key = key.decode().lower()
            if lower_key not in multi_value_headers:
                multi_value_headers[lower_key] = []
            multi_value_headers[lower_key].append(value.decode())

        headers = case_mutated_headers(multi_value_headers)

        body, is_base64_encoded = self._handle_base64_response_body(
            response.body, headers
        )

        out = {
            "statusCode": response.status,
            "body": body,
            "isBase64Encoded": is_base64_encoded,
        }

        #  "You must use multiValueHeaders if you have enabled multi-value headers
        #  and headers otherwise"
        #  https://docs.aws.amazon.com/elasticloadbalancing/latest/application/lambda-functions.html
        multi_value_headers_enabled = "multiValueHeaders" in self.trigger_event
        if multi_value_headers_enabled:
            out["multiValueHeaders"] = multi_value_headers
        else:
            out["headers"] = headers

        return out



from typing import Any
from urllib.parse import urlparse

import aiosqlite

from .base import WebSocketBackend
from ..exceptions import WebSocketError


class SQLiteBackend(WebSocketBackend):
    async def __aenter__(self) -> WebSocketBackend:
        parsed_dsn = urlparse(self.dsn)
        self.connection = await aiosqlite.connect(parsed_dsn.path)
        await self.connection.execute(
            "create table if not exists mangum_websockets "
            "(id varchar(64) primary key, initial_scope text)"
        )
        await self.connection.commit()

        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        await self.connection.close()

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        await self.connection.execute(
            "insert into mangum_websockets values (?, ?)",
            (connection_id, json_scope),
        )
        await self.connection.commit()

    async def retrieve(self, connection_id: str) -> str:
        async with self.connection.execute(
            "select initial_scope from mangum_websockets where id = ?",
            (connection_id,),
        ) as cursor:
            row = await cursor.fetchone()
            if not row:
                raise WebSocketError(f"Connection not found: {connection_id}")
            scope = row[0]

        return scope

    async def delete(self, connection_id: str) -> None:
        await self.connection.execute(
            "delete from mangum_websockets where id = ?", (connection_id,)
        )
        await self.connection.commit()



from typing import Any
import aioredis

from .base import WebSocketBackend
from ..exceptions import WebSocketError


class RedisBackend(WebSocketBackend):
    async def __aenter__(self) -> WebSocketBackend:
        self.connection = await aioredis.create_redis(self.dsn)

        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        self.connection.close()

    async def save(self, connection_id: str, *, json_scope: str) -> None:
        await self.connection.set(connection_id, json_scope)

    async def retrieve(self, connection_id: str) -> str:
        scope = await self.connection.get(connection_id)
        if not scope:
            raise WebSocketError(f"Connection not found: {connection_id}")
        return scope

    async def delete(self, connection_id: str) -> None:
        await self.connection.delete(connection_id)



from .abstract_handler import AbstractHandler
from .aws_alb import AwsAlb
from .aws_api_gateway import AwsApiGateway
from .aws_cf_lambda_at_edge import AwsCfLambdaAtEdge
from .aws_http_gateway import AwsHttpGateway
from .aws_ws_gateway import AwsWsGateway

__all__ = [
    "AbstractHandler",
    "AwsAlb",
    "AwsApiGateway",
    "AwsCfLambdaAtEdge",
    "AwsHttpGateway",
    "AwsWsGateway",
]



import enum
import asyncio
from typing import Optional
import logging
from io import BytesIO
from dataclasses import dataclass

from .. import Response, Request
from ..types import ASGIApp, Message
from ..exceptions import UnexpectedMessage


class HTTPCycleState(enum.Enum):
    """
    The state of the ASGI `http` connection.

    * **REQUEST** - Initial state. The ASGI application instance will be run with the
    connection scope containing the `http` type.
    * **RESPONSE** - The `http.response.start` event has been sent by the application.
    The next expected message is the `http.response.body` event, containing the body
    content. An application may pass the `more_body` argument to send content in chunks,
    however content will always be returned in a single response, never streamed.
    * **COMPLETE** - The body content from the ASGI application has been completely
    read. A disconnect event will be sent to the application, and the response will
    be returned.
    """

    REQUEST = enum.auto()
    RESPONSE = enum.auto()
    COMPLETE = enum.auto()


@dataclass
class HTTPCycle:
    """
    Manages the application cycle for an ASGI `http` connection.

    * **request** - A request object containing the event and context for the connection
    scope used to run the ASGI application instance.
    * **state** - An enumerated `HTTPCycleState` type that indicates the state of the
    ASGI connection.
    * **app_queue** - An asyncio queue (FIFO) containing messages to be received by the
    application.
    * **response** - A dictionary containing the response data to return in AWS Lambda.
    """

    request: Request
    state: HTTPCycleState = HTTPCycleState.REQUEST
    response: Optional[Response] = None

    def __post_init__(self) -> None:
        self.logger: logging.Logger = logging.getLogger("mangum.http")
        self.loop = asyncio.get_event_loop()
        self.app_queue: asyncio.Queue[Message] = asyncio.Queue()
        self.body: BytesIO = BytesIO()

    def __call__(self, app: ASGIApp, initial_body: bytes) -> Response:
        self.logger.debug("HTTP cycle starting.")
        self.app_queue.put_nowait(
            {"type": "http.request", "body": initial_body, "more_body": False}
        )
        asgi_instance = self.run(app)
        asgi_task = self.loop.create_task(asgi_instance)
        self.loop.run_until_complete(asgi_task)

        if self.response is None:
            # Something really bad happened and we puked before we could get a
            # response out
            self.response = Response(
                status=500,
                body=b"Internal Server Error",
                headers=[[b"content-type", b"text/plain; charset=utf-8"]],
            )
        return self.response

    async def run(self, app: ASGIApp) -> None:
        """
        Calls the application with the `http` connection scope.
        """
        try:
            await app(self.request.scope, self.receive, self.send)
        except BaseException as exc:
            self.logger.error("Exception in 'http' protocol.", exc_info=exc)
            if self.state is HTTPCycleState.REQUEST:
                await self.send(
                    {
                        "type": "http.response.start",
                        "status": 500,
                        "headers": [[b"content-type", b"text/plain; charset=utf-8"]],
                    }
                )
                await self.send(
                    {"type": "http.response.body", "body": b"Internal Server Error"}
                )
            elif self.state is not HTTPCycleState.COMPLETE:
                self.response = Response(
                    status=500,
                    body=b"Internal Server Error",
                    headers=[[b"content-type", b"text/plain; charset=utf-8"]],
                )

    async def receive(self) -> Message:
        """
        Awaited by the application to receive ASGI `http` events.
        """
        return await self.app_queue.get()  # pragma: no cover

    async def send(self, message: Message) -> None:
        """
        Awaited by the application to send ASGI `http` events.
        """
        message_type = message["type"]
        self.logger.info(
            "%s:  '%s' event received from application.", self.state, message_type
        )

        if (
            self.state is HTTPCycleState.REQUEST
            and message_type == "http.response.start"
        ):
            if self.response is None:
                self.response = Response(
                    status=message["status"],
                    headers=message.get("headers", []),
                    body=b"",
                )
            self.state = HTTPCycleState.RESPONSE
        elif (
            self.state is HTTPCycleState.RESPONSE
            and message_type == "http.response.body"
        ):
            body = message.get("body", b"")
            more_body = message.get("more_body", False)

            # The body must be completely read before returning the response.
            self.body.write(body)

            if not more_body and self.response is not None:
                body = self.body.getvalue()
                self.body.close()
                self.response.body = body

                self.state = HTTPCycleState.COMPLETE
                await self.app_queue.put({"type": "http.disconnect"})

        else:
            raise UnexpectedMessage(
                f"{self.state}: Unexpected '{message_type}' event received."
            )
